/// <reference types="node" />
/// <reference types="node" />
/// <reference types="node" />
import { EventEmitter } from 'node:events';
import * as net from 'node:net';
import * as tls from 'node:tls';
import { RedisArgument } from '../RESP/types';
export interface RedisSocketCommonOptions {
    /**
     * Connection Timeout (in milliseconds)
     */
    connectTimeout?: number;
    /**
     * Toggle [`Nagle's algorithm`](https://nodejs.org/api/net.html#net_socket_setnodelay_nodelay)
     */
    noDelay?: boolean;
    /**
     * Toggle [`keep-alive`](https://nodejs.org/api/net.html#net_socket_setkeepalive_enable_initialdelay)
     */
    keepAlive?: number | false;
    /**
     * When the socket closes unexpectedly (without calling `.close()`/`.destroy()`), the client uses `reconnectStrategy` to decide what to do. The following values are supported:
     * 1. `false` -> do not reconnect, close the client and flush the command queue.
     * 2. `number` -> wait for `X` milliseconds before reconnecting.
     * 3. `(retries: number, cause: Error) => false | number | Error` -> `number` is the same as configuring a `number` directly, `Error` is the same as `false`, but with a custom error.
     * Defaults to `retries => Math.min(retries * 50, 500)`
     */
    reconnectStrategy?: false | number | ((retries: number, cause: Error) => false | Error | number);
}
type RedisNetSocketOptions = Partial<net.SocketConnectOpts> & {
    tls?: false;
};
export interface RedisTlsSocketOptions extends tls.ConnectionOptions {
    tls: true;
}
export type RedisSocketOptions = RedisSocketCommonOptions & (RedisNetSocketOptions | RedisTlsSocketOptions);
export type RedisSocketInitiator = () => Promise<void>;
export default class RedisSocket extends EventEmitter {
    private static _initiateOptions;
    private static _isTlsSocket;
    private readonly _initiator;
    private readonly _options;
    private _socket?;
    private _isOpen;
    get isOpen(): boolean;
    private _isReady;
    get isReady(): boolean;
    private _isSocketUnrefed;
    constructor(initiator: RedisSocketInitiator, options?: RedisSocketOptions);
    private _reconnectStrategy;
    private _shouldReconnect;
    connect(): Promise<void>;
    private _connect;
    private _createSocket;
    private _createNetSocket;
    private _createTlsSocket;
    private _onSocketError;
    write(iterator: IterableIterator<Array<RedisArgument>>): void;
    quit<T>(fn: () => Promise<T>): Promise<T>;
    close(): void;
    destroy(): void;
    destroySocket(): void;
    ref(): void;
    unref(): void;
}
export {};
